unit Rectangles;
  { provides ADT for axis-parallel grid rectangles
    with abstract base class and two implementations }

interface

type
  TRectangle = class(TObject)  // axis-parallel grid rectangles
    // abstract base class

  public
    // construction
    constructor Create(AXL, AYL, AXH, AYH: Integer);
      // pre: AXL <= AXH  /\  AYL <= AYH
      // post: XL=AXL /\ YL=AYL /\ XH=AXH /\ YH=AYH
      virtual;

    // basic queries
    function XL: Integer; // lower X coordinate 
      virtual; abstract;
    function YL: Integer; // lower Y coordinate
      virtual; abstract;
    function XH: Integer; // higher X coordinate
      virtual; abstract;
    function YH: Integer; // higher Y coordinate
      virtual; abstract;

    // invariants
    // WellFormed: XL <= XH  /\  YL <= YH

    // derived queries
    function ToStr: String;
      // pre: true
      // ret: string representation of Self
      virtual;
    function Contains(AX, AY: Integer): Boolean;
      // pre: true
      // ret: XL <= AX <= XH  /\  YL <= AY <= YH
      virtual;
    function Intersects(const ARectangle: TRectangle): Boolean;
      // pre: true
      // ret: Self intersects ARectangle
      virtual;

    //commands
    procedure SetXL(AX: Integer);
      // pre: AX <= XH
      // effect: XL := AX
      virtual; abstract;
    procedure SetYL(AY: Integer);
      // pre: AY <= YH
      // effect: YL := AY
      virtual; abstract;
    procedure SetXH(AX: Integer);
      // pre: XL <= AX
      // effect: XH := AX
      virtual; abstract;
    procedure SetYH(AY: Integer);
      // pre: YL <= AY
      // effect: YH := AY
      virtual; abstract;
    procedure Shift(AX, AY: Integer);
      // pre: true
      // effect: XL, YL, XH, YL := XL+AX, YL+AY, XH+AX, YH+AY
      virtual; abstract;
    procedure Intersect(const ARectangle: TRectangle);
      // pre: Intersects(ARectangle)
      // effect: Self := Self intersection ARectangle
      virtual;
    procedure Hull(const ARectangle: TRectangle);
      // pre: true
      // effect: Self := smallest rectangle enclosing
      //                 Self and ARectangle
      virtual;
  end; { class TRectangle }

  TRectangleLowHigh = class(TRectangle)
    // implementation storing low/high x/y coordinates
  private
    // fields
    FXL: Integer; // lower X coordinate
    FYL: Integer; // lower Y coordinate
    FXH: Integer; // higher X coordinate
    FYH: Integer; // higher Y coordinate

  public
    // construction
    constructor Create(AXL, AYL, AXH, AYH: Integer); override;

    // basic queries
    function XL: Integer; override;
    function YL: Integer; override;
    function XH: Integer; override;
    function YH: Integer; override;

    //commands
    procedure SetXL(AX: Integer); override;
    procedure SetYL(AY: Integer); override;
    procedure SetXH(AX: Integer); override;
    procedure SetYH(AY: Integer); override;
    procedure Shift(AX, AY: Integer); override;
  end; { class TRectangleLowHigh }

  TRectangleLowDim = class(TRectangle)
    // implementation storing low x/y and dimensions
  private
    // fields
    FXL: Integer; // lower X coordinate
    FYL: Integer; // lower Y coordinate
    FWidth: Integer; // width                       (*CHANGED*)
    FHeight: Integer; // height                     (*CHANGED*)

    // invariants                                     (*ADDED*)
    // Positive: 1 <= FWidth  /\  1 <= FHeight        (*ADDED*)
    // HighCoord: XH = FXL + FWidth - 1  /\           (*ADDED*)
    //            YH = FYL + FHeight - 1              (*ADDED*)

  public
    // construction
    constructor Create(AXL, AYL, AXH, AYH: Integer); override;

    // basic queries
    function XL: Integer; override;
    function YL: Integer; override;
    function XH: Integer; override;
    function YH: Integer; override;

    //commands
    procedure SetXL(AX: Integer); override;
    procedure SetYL(AY: Integer); override;
    procedure SetXH(AX: Integer); override;
    procedure SetYH(AY: Integer); override;
    procedure Shift(AX, AY: Integer); override;
  end; { class TRectangleLowDim }

implementation

uses
  SysUtils, { for Format }
  Math; { for Min and Max }

// Implementation of TRectangle ---------------------------

constructor TRectangle.Create(AXL, AYL, AXH, AYH: Integer);
  begin
    // check pre
    Assert((AXL <= AXH) and (AYL <= AYH),
      Format('TRectangle.Create(%D, %D, %D, %D).pre failed',
             [AXL, AYL, AXH, AYH]))
  ; inherited Create
  end; { Create }

function TRectangle.ToStr: String;
  begin
    Result := Format('(%D, %D; %D, %D)', [XL, YL, XH, YH])
  end; { ToStr }

function TRectangle.Contains(AX, AY: Integer): Boolean;
  begin
    Result := (XL <= AX) and (AX <= XH) and
              (YL <= AY) and (AY <= YH)
  end; { Contains }

function TRectangle.Intersects(const ARectangle: TRectangle): Boolean;
  begin
    Result := (XL <= ARectangle.XH)  // not Self right of ARectangle
          and (ARectangle.XL <= XH)  // not Self left of ARectangle
          and (YL <= ARectangle.YH)  // not Self above ARectangle
          and (ARectangle.YL <= YH)  // not Self below ARectangle
  end; { Intersects }

procedure TRectangle.Intersect(const ARectangle: TRectangle);
  begin
    // check pre
    Assert(Intersects(ARectangle),
      Format('TRectangle.Intersect(%S).pre failed: Self = %S',
             [ARectangle.ToStr, ToStr]))
  ; SetXL(Max(XL, ARectangle.XL))
  ; SetYL(Max(YL, ARectangle.YL))
  ; SetXH(Min(XH, ARectangle.XH))
  ; SetYH(Min(YH, ARectangle.YH))
  end; { Intersect }

procedure TRectangle.Hull(const ARectangle: TRectangle);
  begin
    SetXL(Min(XL, ARectangle.XL))
  ; SetYL(Min(YL, ARectangle.YL))
  ; SetXH(Max(XH, ARectangle.XH))
  ; SetYH(Max(YH, ARectangle.YH))
  end; { Hull }

// Implementation of TRectangleLowHigh --------------------

constructor TRectangleLowHigh.Create(AXL, AYL, AXH, AYH: Integer);
  begin
  ; inherited Create(AXL, AYL, AXH, AYH)
  ; FXL := AXL
  ; FYL := AYL
  ; FXH := AXH
  ; FYH := AYH
  end; { Create }

function TRectangleLowHigh.XL: Integer;
  begin
    Result := FXL
  end; { XL }

function TRectangleLowHigh.YL: Integer;
  begin
    Result := FYL
  end; { YL }

function TRectangleLowHigh.XH: Integer;
  begin
    Result := FXH
  end; { XH }

function TRectangleLowHigh.YH: Integer;
  begin
    Result := FYH
  end; { YH }

procedure TRectangleLowHigh.SetXL(AX: Integer);
  begin
    // check pre
    Assert(AX <= XH,
      Format('TRectangleLowHigh.SetXL(%D).pre failed: XH = %D',
             [AX, XH]))
  ; FXL := AX
  end; { SetXL }

procedure TRectangleLowHigh.SetYL(AY: Integer);
  begin
    // check pre
    Assert(AY <= YH,
      Format('TRectangleLowHigh.SetYL(%D).pre failed: YH = %D',
             [AY, YH]))
  ; FYL := AY
  end; { SetYL }

procedure TRectangleLowHigh.SetXH(AX: Integer);
  begin
    // check pre
    Assert(XL <= AX,
      Format('TRectangleLowHigh.SetXH(%D).pre failed: XL = %D',
             [AX, XL]))
  ; FXH := AX
  end; { SetXH }

procedure TRectangleLowHigh.SetYH(AY: Integer);
  begin
    // check pre
    Assert(YL <= AY,
      Format('TRectangleLowHigh.SetYH(%D).pre failed: YL = %D',
             [AY, YL]))
  ; FYH := AY
  end; { SetYH }

procedure TRectangleLowHigh.Shift(AX, AY: Integer);
  begin
    FXL := FXL + AX
  ; FYL := FYL + AY
  ; FXH := FXH + AX
  ; FYH := FYH + AY
  end; { Shift }

// Implementation of TRectangleLowDim ---------------------

constructor TRectangleLowDim.Create(AXL, AYL, AXH, AYH: Integer);
  begin
  ; inherited Create(AXL, AYL, AXH, AYH)
  ; FXL := AXL
  ; FYL := AYL
  ; FWidth := AXH - AXL + 1                         (*CHANGED*)
  ; FHeight := AYH - AYL + 1                        (*CHANGED*)
  end; { Create }

function TRectangleLowDim.XL: Integer;
  begin
    Result := FXL
  end; { XL }

function TRectangleLowDim.YL: Integer;
  begin
    Result := FYL
  end; { YL }

function TRectangleLowDim.XH: Integer;
  begin
    Result := FXL + FWidth - 1                      (*CHANGED*)
  end; { XH }

function TRectangleLowDim.YH: Integer;
  begin
    Result := FYL + FHeight - 1                     (*CHANGED*)
  end; { YH }

procedure TRectangleLowDim.SetXL(AX: Integer);
  begin
    // check pre
    Assert(AX <= XH,
      Format('TRectangleLowDim.SetXL(%D).pre failed: XH = %D',
             [AX, XH]))
  ; FWidth := XH - AX + 1  { must be done first }     (*ADDED*)
  ; FXL := AX
  end; { SetXL }

procedure TRectangleLowDim.SetYL(AY: Integer);
  begin
    // check pre
    Assert(AY <= YH,
      Format('TRectangleLowDim.SetYL(%D).pre failed: YH = %D',
             [AY, YH]))
  ; FHeight := YH - AY + 1  { must be done first }    (*ADDED*)
  ; FYL := AY
  end; { SetYL }

procedure TRectangleLowDim.SetXH(AX: Integer);
  begin
    // check pre
    Assert(XL <= AX,
      Format('TRectangleLowDim.SetXH(%D).pre failed: XL = %D',
             [AX, XL]))
  ; FWidth := AX - FXL + 1                          (*CHANGED*)
  end; { SetXH }

procedure TRectangleLowDim.SetYH(AY: Integer);
  begin
    // check pre
    Assert(YL <= AY,
      Format('TRectangleLowDim.SetYH(%D).pre failed: YL = %D',
             [AY, YL]))
  ; FHeight := AY - FYL + 1                         (*CHANGED*)
  end; { SetYH }

procedure TRectangleLowDim.Shift(AX, AY: Integer);
  begin
    FXL := FXL + AX
  ; FYL := FYL + AY
                                                    (*DELETED*)
                                                    (*DELETED*)
  end; { Shift }

end.
